from tqdm import tqdm
import torch
import torch.nn.functional as F
from torchvision import datasets, transforms
import matplotlib.pyplot as plt
import torchattacks


def train(model, criterion, device, num_classes, train_loader, optimizer, epoch, noise_sd=None):
    model.train()

    #pbar = tqdm(total=len(train_loader), position=0, leave=True)
    with tqdm(total=len(train_loader), position=0, leave=True) as pbar:
        for batch_idx, (data, target) in enumerate(train_loader, start=1):
            #data = data.view(data.shape[0], -1)
            data, target = data.to(device), target.to(device)

            if noise_sd is not None:
                data = data + torch.randn_like(data, device='cuda') * noise_sd

            #         data.requires_grad = True
            optimizer.zero_grad()
            out = model(data)
            if str(criterion) == 'CrossEntropyLoss()':
                loss = criterion(out, target)
            elif str(criterion) == 'MSELoss()':
                loss = criterion(out, F.one_hot(target, num_classes=num_classes).float())

            loss.backward()
            optimizer.step()

            accuracy = torch.mean((torch.argmax(out, dim=1) == target).float()).item()

            pbar.update(1)
            pbar.set_description(
                'Epoch: {} [{}/{} ({:.0f}%)] | Batch Loss: {:.6f} | Batch Accuracy: {:.6f}'.format(
                    epoch,
                    batch_idx,
                    len(train_loader),
                    100. * batch_idx / len(train_loader),
                    loss.item(),
                    accuracy))

        #         if debug and batch_idx > 20:
        #             break
    #pbar.close()


def test(model, device, test_loader, noise_sd=None):
    model.eval()
    n_crr = 0
    size = 0
    #     pbar = tqdm(total=len(test_loader), position=0, leave=True)
    for batch_idx, (data, target) in enumerate(test_loader, start=1):
        #data = data.view(data.shape[0], -1)
        #         if data.shape[0] != batch_size:
        #             continue

        data, target = data.to(device), target.to(device)

        if noise_sd is not None:
            data = data + torch.randn_like(data, device='cuda') * noise_sd

        out = model(data)

        n_crr += torch.sum((torch.argmax(out, dim=1) == target).float()).item()
        size += data.shape[0]
    acc = n_crr / size
    print(acc)


def train_robust_data(model, criterion, device, train_loader, robust_train_loader, optimizer,
                      epoch, attack_type, alpha0=1, beta=1, eps=0.3, step_size=0.02, steps=10, noise_sd=None):
    model.train()
    total_loss = 0
    avg_loss = AverageMeter()
    avg_acc = AverageMeter()
    pbar = tqdm(total=len(train_loader), position=0, leave=True)
    beta = beta / (2 ** (int(epoch / 30)))

    if attack_type == 'inf':
        atk = torchattacks.PGD(model, eps=eps, alpha=step_size, steps=steps)
    elif attack_type == 'l2':
        atk = torchattacks.PGDL2(model, eps=eps, alpha=step_size, steps=steps)
    else:
        raise NotImplementedError

    for batch_idx, ((data, target), (robust_data, robust_target)) in enumerate(zip(train_loader, robust_train_loader),
                                                                               start=1):
        #         data=data.view(data.shape[0],-1)
        #         robust_data=robust_data.view(robust_data.shape[0],-1)
        data, target = data.to(device), target.to(device)
        robust_data, robust_target = robust_data.to(device), robust_target.to(device)

        # augment inputs with noise
        if noise_sd is not None:
            data = data + torch.randn_like(data, device='cuda') * noise_sd
            #robust_data = robust_data + torch.randn_like(robust_data, device='cuda') * noise_sd


        robust_data.requires_grad = True

        #         if epoch>5:
        optimizer.zero_grad()
        #         print("1",torch.cuda.memory_allocated())
        if noise_sd is not None:
            out = model(robust_data + torch.randn_like(robust_data, device='cuda') * noise_sd)
        else:
            out = model(robust_data)

        loss = criterion(out, robust_target)

        loss.backward(retain_graph=True, create_graph=True)
        robust_data.grad = None
        #         print(model.linear1.weight.grad)
        #         print("2",torch.cuda.memory_allocated())
        model.fill_gd(alpha=alpha0)
        #         data.requires_grad = True
        model.eval()
        adv_images = atk(data, target).to(device)
        #         adv_images.grad=None
        #         robust_data.grad=None
        #         adv_images.detach()
        #         out_nat = model(data)
        #         loss_nat = criterion(out_nat, target)
        #         gd_nat  = torch.autograd.grad(loss_nat, data, grad_outputs=None, only_inputs=True, retain_graph=True)[0]
        #         adv_images =  torch.clamp(data + eps*torch.sign(gd_nat), min=0, max=1)
        #         print("3",torch.cuda.memory_allocated())
        #         data.grad=None
        model.train()
        out_adv = model(adv_images)
        loss_adv = criterion(out_adv, target)

        gd_robust = torch.autograd.grad(loss_adv, robust_data, grad_outputs=None, only_inputs=True, retain_graph=False)[
            0]
        robust_data = torch.clamp(robust_data - beta * torch.sign(gd_robust), min=0, max=1)
        #         robust_data = torch.clamp(robust_data-beta*gd_robust, min=0, max=1)
        accuracy = torch.mean((torch.argmax(out_adv, dim=1) == target).float()).item()
        #         print("4",torch.cuda.memory_allocated())
        #         print(robust_data[0].shape)
        if batch_idx == 1:
            robust_dataset = robust_data.detach().cpu()
            fig = plt.figure
            plt.imshow(transforms.ToPILImage()(robust_data[0].detach().cpu()))
            #             plt.imshow(robust_data[0].detach().cpu().view(32,32).numpy(), cmap='gray')
            plt.show()
            robust_label = robust_target.detach().cpu()
        else:
            robust_dataset = torch.cat((robust_dataset, robust_data.detach().cpu()), dim=0)
            robust_label = torch.cat((robust_label, robust_target.detach().cpu()), dim=0)
        #         print(robust_dataset.shape)
        total_loss = total_loss + loss_adv
        avg_loss.update(loss_adv.item())
        avg_acc.update(accuracy)
        #         model.update_gd(alpha= alpha)
        #         print(model.feature_extractor.conv1.weight.grad)
        model.kill_gd()
        #         print(model.feature_extractor.conv1.weight_grad)
        #         print("5",torch.cuda.memory_allocated())
        optimizer.step()
        #         del out_adv,out_nat,out,loss,loss_nat,loss_adv,robust_data,data,adv_img,gd_robust,gd_nat,target,robust_target
        #         print("6",torch.cuda.memory_allocated())
        pbar.update(1)
        pbar.set_description(
            'Epoch: {} [{}/{} ({:.0f}%)] Avg Loss: {:.6f} Avg Accuracy: {:.6f}'.format(
                epoch,
                batch_idx,
                len(train_loader),
                100. * batch_idx / len(train_loader),
                avg_loss.mean,
                avg_acc.mean))

    pbar.close()
    #     del model
    #     print("7",torch.cuda.memory_allocated())
    print(total_loss)
    return robust_dataset, robust_label


def attack(model, test_loader, device, attack_type, eps=16 / 255, alpha=4 / 255, steps=5):
    #     data = []
    model.eval()
    size = 0
    n_crr = 0
    if attack_type == 'inf':
        atk = torchattacks.PGD(model, eps=eps, alpha=alpha, steps=steps)
    elif attack_type == 'l2':
        atk = torchattacks.PGDL2(model, eps=eps, alpha=alpha, steps=steps)
    elif attack_type == 'auto_inf':
        atk = torchattacks.AutoAttack(model, norm='Linf', eps=eps, n_classes=200)
    elif attack_type == 'auto_l2':
        atk = torchattacks.AutoAttack(model, norm='L2', eps=eps, n_classes=200)
    else:
        raise NotImplementedError

    for i, (data, target) in enumerate(tqdm(test_loader)):
        #         print(data.shape)
        #         data=data.view(data.shape[0],-1)
        adv_images = atk(data, target).to(device)
        #         print(adv_images.shape)
        #         data.append(X)
        target = target.to(device)
        outputs = model(adv_images)

        n_crr += torch.sum((torch.argmax(outputs, dim=1) == target).float()).item()
        size += data.shape[0]
    print(n_crr / size)


def adv_data(model, criterion, device, num_classes, train_loader, optimizer, epoch, attack_type, eps=16 / 255, alpha=16 / 255,
             steps=1):
    model.eval()

    if attack_type == 'inf':
        atk = torchattacks.PGD(model, eps=eps, alpha=alpha, steps=steps)
    elif attack_type == 'l2':
        atk = torchattacks.PGDL2(model, eps=eps, alpha=alpha, steps=steps)
    elif attack_type == 'auto_inf':
        atk = torchattacks.AutoAttack(model, norm='Linf', eps=eps, n_classes=200)
    elif attack_type == 'auto_l2':
        atk = torchattacks.AutoAttack(model, norm='L2', eps=eps, n_classes=200)
    else:
        raise NotImplementedError

    for batch_idx, (data, target) in enumerate(tqdm(train_loader), start=1):
        #         data=data.view(data.shape[0],-1)

        data, target = data.to(device), target.to(device)
        adv_images = atk(data, target).to(device)
        if batch_idx == 1:
            robust_dataset = adv_images.detach().cpu()
            #             fig = plt.figure
            #             plt.imshow(adv_images[0].detach().cpu().view(32,32).numpy(), cmap='gray')
            #             plt.show()
            robust_label = target.detach().cpu()
        else:
            robust_dataset = torch.cat((robust_dataset, adv_images.detach().cpu()), dim=0)
            robust_label = torch.cat((robust_label, target.detach().cpu()), dim=0)

    return robust_dataset, robust_label


class AverageMeter(object):
    """Computes and stores the average and current value"""
    def __init__(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def reset(self):
        self.__init__()

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    @property
    def mean(self):
        try:
            return self.sum / self.count
        except Exception:
            return -1
